function Init()
    indicator:name("Daily Pivot");
    indicator:description("Daily Pivot with customizable begin of trading day. Can be applied on H1 or smaller time frame");
    indicator:requiredSource(core.Bar);
    indicator:type(core.Indicator);
    indicator:setTag("group", "Support/Resistance");

    indicator.parameters:addGroup("Pivot Parameters");
    indicator.parameters:addInteger("FH", "Start Hour", "Start hour of the trading day in EST time zone (0-23)", 17, 0, 23);

    indicator.parameters:addString("CalcMode", "Calculate Mode", "", "Pivot");
    indicator.parameters:addStringAlternative("CalcMode", "Classic Pivot", "", "Pivot");
    indicator.parameters:addStringAlternative("CalcMode", "Camarilla", "", "Camarilla");
    indicator.parameters:addStringAlternative("CalcMode", "Woodie", "", "Woodie");
    indicator.parameters:addStringAlternative("CalcMode", "Fibonacci Pivot", "", "Fibonacci");
    indicator.parameters:addStringAlternative("CalcMode", "Floor Pivot", "", "Floor");
    indicator.parameters:addStringAlternative("CalcMode", "Fibonacci Retracement", "", "FibonacciR");

    indicator.parameters:addGroup("Style");
    indicator.parameters:addString("ShowMode", "Display Mode", "", "TODAY");
    indicator.parameters:addStringAlternative("ShowMode", "Today Only", "", "TODAY");
    indicator.parameters:addStringAlternative("ShowMode", "Historical", "", "HIST");
    indicator.parameters:addString("LabelLoc", "Label Location for today mode", "", "E");
    indicator.parameters:addStringAlternative("LabelLoc", "Line End", "", "E");
    indicator.parameters:addStringAlternative("LabelLoc", "Line Start", "", "B");
    indicator.parameters:addStringAlternative("LabelLoc", "Both Sides", "", "A");

    indicator.parameters:addColor("clrP", "Color of P Line", "", core.rgb(192, 192, 192));
    indicator.parameters:addInteger("widthP", "Width of P Line", "", 1, 1, 5);
    indicator.parameters:addInteger("styleP", "Style of P Line", "", core.LINE_SOLID);
    indicator.parameters:setFlag("styleP", core.FLAG_LEVEL_STYLE);

    indicator.parameters:addColor("clrS1", "Color of S1 Line", "", core.rgb(255, 0, 0));
    indicator.parameters:addInteger("widthS1", "Width of S1 Line", "", 1, 1, 5);
    indicator.parameters:addInteger("styleS1", "Style of S1 Line", "", core.LINE_SOLID);
    indicator.parameters:setFlag("styleS1", core.FLAG_LEVEL_STYLE);
    indicator.parameters:addBoolean("showS1", "Show S1 Line", "", true);

    indicator.parameters:addColor("clrS2", "Color of S2 Line", "", core.rgb(224, 0, 0));
    indicator.parameters:addInteger("widthS2", "Width of S2 Line", "", 1, 1, 5);
    indicator.parameters:addInteger("styleS2", "Style of S2 Line", "", core.LINE_SOLID);
    indicator.parameters:setFlag("styleS2", core.FLAG_LEVEL_STYLE);
    indicator.parameters:addBoolean("showS2", "Show S2 Line", "", true);

    indicator.parameters:addColor("clrS3", "Color of S3 Line", "", core.rgb(192, 0, 0));
    indicator.parameters:addInteger("widthS3", "Width of S3 Line", "", 1, 1, 5);
    indicator.parameters:addInteger("styleS3", "Style of S3 Line", "", core.LINE_SOLID);
    indicator.parameters:setFlag("styleS3", core.FLAG_LEVEL_STYLE);
    indicator.parameters:addBoolean("showS3", "Show S3 Line", "", true);

    indicator.parameters:addColor("clrS4", "Color of S4 Line", "", core.rgb(160, 0, 0));
    indicator.parameters:addInteger("widthS4", "Width of S4 Line", "", 1, 1, 5);
    indicator.parameters:addInteger("styleS4", "Style of S4 Line", "", core.LINE_SOLID);
    indicator.parameters:setFlag("styleS4", core.FLAG_LEVEL_STYLE);
    indicator.parameters:addBoolean("showS4", "Show S4 Line", "", true);

    indicator.parameters:addColor("clrR1", "Color of R1 Line", "", core.rgb(0, 255, 0));
    indicator.parameters:addInteger("widthR1", "Width of R1 Line", "", 1, 1, 5);
    indicator.parameters:addInteger("styleR1", "Style of R1 Line", "", core.LINE_SOLID);
    indicator.parameters:setFlag("styleR1", core.FLAG_LEVEL_STYLE);
    indicator.parameters:addBoolean("showR1", "Show R1 Line", "", true);

    indicator.parameters:addColor("clrR2", "Color of R2 Line", "", core.rgb(0, 224, 0));
    indicator.parameters:addInteger("widthR2", "Width of R2 Line", "", 1, 1, 5);
    indicator.parameters:addInteger("styleR2", "Style of R2 Line", "", core.LINE_SOLID);
    indicator.parameters:setFlag("styleR2", core.FLAG_LEVEL_STYLE);
    indicator.parameters:addBoolean("showR2", "Show R2 Line", "", true);

    indicator.parameters:addColor("clrR3", "Color of R3 Line", "", core.rgb(0, 192, 0));
    indicator.parameters:addInteger("widthR3", "Width of R3 Line", "", 1, 1, 5);
    indicator.parameters:addInteger("styleR3", "Style of R3 Line", "", core.LINE_SOLID);
    indicator.parameters:setFlag("styleR3", core.FLAG_LEVEL_STYLE);
    indicator.parameters:addBoolean("showR3", "Show R3 Line", "", true);

    indicator.parameters:addColor("clrR4", "Color of R4 Line", "", core.rgb(0, 160, 0));
    indicator.parameters:addInteger("widthR4", "Width of R4 Line", "", 1, 1, 5);
    indicator.parameters:addInteger("styleR4", "Style of R4 Line", "", core.LINE_SOLID);
    indicator.parameters:setFlag("styleR4", core.FLAG_LEVEL_STYLE);
    indicator.parameters:addBoolean("showR4", "Show R4 Line", "", true);

    indicator.parameters:addColor("clrMP", "Color of Midpoint line", "", core.rgb(128, 128, 128));
    indicator.parameters:addInteger("widthMP", "Width of Midpoint Line", "", 1, 1, 5);
    indicator.parameters:addInteger("styleMP", "Style of Midpoint Line", "", core.LINE_DOT);
    indicator.parameters:setFlag("styleMP", core.FLAG_LEVEL_STYLE);
    indicator.parameters:addBoolean("ShowMP", "Show Midpoint Line", "", false);
end

local P;
local H;
local L;
local T;
local C;
local D;
local source;
local ref;
local instr;
local BS;
local CurrLen;
local BSLen;
local host;
local offset;

local RP = 0;
local S1 = 1;
local S2 = 2;
local S3 = 3;
local S4 = 4;
local R1 = 5;
local R2 = 6;
local R3 = 7;
local R4 = 8;
local PID = 9;
local name = {};
local show = {};
local clr = {};
local width = {};
local style = {};
local stream = {};
local fibr = {};

local clrP;
local widthP;
local styleP;

local O_PIVOT = 1;
local O_CAM = 2;
local O_WOOD = 3;
local O_FIB = 4;
local O_FLOOR = 5;
local O_FIBR = 6;
local CalcMode;

local O_HIST = 1;
local O_TODAY = 2;
local ShowMode;
local ShowMP;
local clrMP;
local widthMP;
local styleMP;
local O_END = 1;
local O_BEG = 2;
local O_BOTH = 3;
local LabelLoc;
local eps;
local firstHour, length;

function Prepare(onlyName)
    host = core.host;
    offset = host:execute("getTradingDayOffset");

    source = instance.source;
    instr = source:instrument();
    BS = instance.parameters.BS;
    clrP = instance.parameters.clrP;
    widthP = instance.parameters.widthP;
    styleP = instance.parameters.styleP;
    ShowMP = instance.parameters.ShowMP;
    clrMP = instance.parameters.clrMP;
    widthMP = instance.parameters.widthMP;
    styleMP = instance.parameters.styleMP;

    local precision = source:getPrecision();
    if precision > 0 then
        eps = math.pow(10, -precision);
    else
        eps = 1;
    end

    name[RP] = "P";
    name[S1] = "S1";
    name[S2] = "S2";
    name[S3] = "S3";
    name[S4] = "S4";
    name[R1] = "R1";
    name[R2] = "R2";
    name[R3] = "R3";
    name[R4] = "R4";
    show[S1] = instance.parameters.showS1;
    show[S2] = instance.parameters.showS2;
    show[S3] = instance.parameters.showS3;
    show[S4] = instance.parameters.showS4;
    show[R1] = instance.parameters.showR1;
    show[R2] = instance.parameters.showR2;
    show[R3] = instance.parameters.showR3;
    show[R4] = instance.parameters.showR4;
    clr[S1] = instance.parameters.clrS1;
    clr[S2] = instance.parameters.clrS2;
    clr[S3] = instance.parameters.clrS3;
    clr[S4] = instance.parameters.clrS4;
    clr[R1] = instance.parameters.clrR1;
    clr[R2] = instance.parameters.clrR2;
    clr[R3] = instance.parameters.clrR3;
    clr[R4] = instance.parameters.clrR4;
    width[S1] = instance.parameters.widthS1;
    width[S2] = instance.parameters.widthS2;
    width[S3] = instance.parameters.widthS3;
    width[S4] = instance.parameters.widthS4;
    width[R1] = instance.parameters.widthR1;
    width[R2] = instance.parameters.widthR2;
    width[R3] = instance.parameters.widthR3;
    width[R4] = instance.parameters.widthR4;
    style[S1] = instance.parameters.styleS1;
    style[S2] = instance.parameters.styleS2;
    style[S3] = instance.parameters.styleS3;
    style[S4] = instance.parameters.styleS4;
    style[R1] = instance.parameters.styleR1;
    style[R2] = instance.parameters.styleR2;
    style[R3] = instance.parameters.styleR3;
    style[R4] = instance.parameters.styleR4;

    fibr[S4] = -0.272;
    fibr[S3] = 0;
    fibr[S2] = 0.236;
    fibr[S1] = 0.382;
    fibr[R1] = 0.618;
    fibr[R2] = 0.764;
    fibr[R3] = 1;
    fibr[R4] = 1.272;

    -- validate
    local l1, l2;
    local s, e;

    s, e = core.getcandle(source:barSize(), core.now(), 0);
    l1 = math.floor((e - s) * 86400.0 + 0.5) / 86400.0;
    CurrLen = l1;

    firstHour = instance.parameters.FH;
    if firstHour >= 12 then
        firstHour = firstHour - 24;
    end

    if instance.parameters.ShowMode == "TODAY" then
        ShowMode = O_TODAY;
    elseif instance.parameters.ShowMode == "HIST" then
        ShowMode = O_HIST;
    else
        assert(false, "Unkown display mode" .. ": " .. instance.parameters.CalcMode);
    end

    if instance.parameters.CalcMode == "Pivot" then
        CalcMode = O_PIVOT;
    elseif instance.parameters.CalcMode == "Camarilla" then
        CalcMode = O_CAM;
    elseif instance.parameters.CalcMode == "Woodie" then
        CalcMode = O_WOOD;
    elseif instance.parameters.CalcMode == "Fibonacci" then
        CalcMode = O_FIB;
    elseif instance.parameters.CalcMode == "Floor" then
        CalcMode = O_FLOOR;
    elseif instance.parameters.CalcMode == "FibonacciR" then
        CalcMode = O_FIBR;
        if ShowMode == O_TODAY then
            name[S1] = tostring(fibr[S1]);
            name[S2] = tostring(fibr[S2]);
            name[S3] = tostring(fibr[S3]);
            name[S4] = tostring(fibr[S4]);
            name[R1] = tostring(fibr[R1]);
            name[R2] = tostring(fibr[R2]);
            name[R3] = tostring(fibr[R3]);
            name[R4] = tostring(fibr[R4]);
            name[RP] = "0.5";
        end
    else
        assert(false, "Unknown calculation mode" .. ": " .. instance.parameters.CalcMode);
    end


    if instance.parameters.LabelLoc == "E" then
        LabelLoc = O_END;
    elseif instance.parameters.LabelLoc == "B" then
        LabelLoc = O_BEG;
    elseif instance.parameters.LabelLoc == "A" then
        LabelLoc = O_BOTH;
    else
        assert(false, "Unknown label location" .. ": " .. instance.parameters.LabelLoc);
    end

    -- create streams
    local sname;
    sname = profile:id() .. "(" .. source:name() .. "," .. instance.parameters.CalcMode .. "," .. instance.parameters.FH ..":00)";
    instance:name(sname);

    if onlyName then
        assert(l1 <= 1.0 / 24.0, "The indicator can be applied on hour or smaller time frames only");
        return;
    end

    -- pivot
    if ShowMode == O_HIST then
        P = instance:addStream("P", core.Line, sname .. "." .. "P", "P", clrP, 0);
        P:setWidth(widthP);
        P:setStyle(styleP);
    else
        D = instance:addInternalStream(0, 0);
        D:setWidth(widthP);
        D:setStyle(styleP);
        P = instance:addInternalStream(0, 0);
    end
    -- range
    T = instance:addInternalStream(0, 0);
    H = instance:addInternalStream(0, 0);
    L = instance:addInternalStream(0, 0);
    C = instance:addInternalStream(0, 0);
    -- show stream for historical mode
    if ShowMode == O_HIST then
        for i = S1, R4, 1 do
            if show[i] then
                stream[i] = instance:addStream(name[i], core.Line, sname .. "." .. name[i], name[i], clr[i], 0);
                stream[i]:setWidth(width[i]);
                stream[i]:setStyle(style[i]);
            end
        end
    end
end

local loading = false;
local loadingFrom, loadingTo;
local pday = nil;
local d = {};
local canWork = nil;

function Update(period, mode)
    if canWork == nil then
        if CurrLen > 1.0 / 24.0 then
            core.host:execute("setStatus", "The indicator can be applied on hour or smaller time frames only");
            canWork = false;
            return ;
        else
            canWork = true;
        end
    elseif not(canWork) then
        return ;
    end

    -- get the previous's candle and load the ref data in case ref data does not exist
    local candle, candle_end;
    candle, candle_end = getPrevCandle(source:date(period));

    -- if data for the specific candle are still loading
    -- then do nothing
    if loading and candle >= loadingFrom and (loadingTo == 0 or candle <= loadingTo) then
        return ;
    end

    -- if data is not loaded yet at all
    -- load the data
    if ref == nil then
        -- there is no data at all, load initial data
        local to, t;
        local from;
        if (source:isAlive()) then
            -- if the source is subscribed for updates
            -- then subscribe the current collection as well
            to = 0;
        else
            -- else load up to the last currently available date
            to = getCandle(source:date(source:size() - 1)) + 1;
        end
        from = getPrevCandle(source:date(source:first()));
        loading = true;
        P:setBookmark(1, period);
        loadingFrom = from;
        loadingTo = to;
        ref = host:execute("getHistory", 1, source:instrument(), "H1", from, to, source:isBid());
        return ;
    end

    if ref:size() == 0 then
        return;
    end
    -- check whether the requested candle is before
    -- the reference collection start
    if (candle < ref:date(0)) then
        local from;
        P:setBookmark(1, period);
        if loading then
            return ;
        end
        loading = true;
        loadingFrom = candle;
        loadingTo = ref:date(0);
        host:execute("extendHistory", 1, ref, loadingFrom, loadingTo);
        return ;
    end

    -- check whether the requested candle is after
    -- the reference collection end
    if (not(source:isAlive()) and candle > ref:date(ref:size() - 1)) then
        if loading then
            return ;
        end
        loading = true;
        P:setBookmark(1, period);
        loadingFrom = ref:date(ref:size() - 1);
        loadingTo = candle;
        host:execute("extendHistory", 1, ref, loadingFrom, loadingTo);
        return ;
    end

    T[period] = -1;
    if period == 0 or T[period - 1] ~= candle then
--if period ~= source:size() - 1 then
--    core.host:trace(string.format("%s %s %i %i", core.formatDate(source:date(period)), core.formatDate(candle), firstHour, offset));
--end
        local start = core.findDate(ref, candle, false);
        if start >= 0 then
            local h, l, c;
            h = 0;
            l = 10^10;
            c = nil;
            while start < ref:size() and ref:date(start) < candle_end do
                if (ref:date(start) >= candle) then
                    h = math.max(h, ref.high[start]);
                    l = math.min(l, ref.low[start]);
                    c = ref.close[start];
                end
                start = start + 1;
            end

            if c ~= nil then
                T[period] = candle;
                H[period] = h;
                L[period] = l;
                C[period] = c;
            else
                T[period] = candle;
                H[period] = H[period - 1];
                L[period] = L[period - 1];
                C[period] = C[period - 1];
            end
        end
    else
        T[period] = candle;
        H[period] = H[period - 1];
        L[period] = L[period - 1];
        C[period] = C[period - 1];
    end

    if T[period] == -1 then
        return ;
    end


    if CalcMode == O_PIVOT or CalcMode == O_FIB or CalcMode == O_FLOOR then
        P[period] = (H[period] + C[period] + L[period]) / 3;
    elseif CalcMode == O_CAM then
        P[period] = C[period];
    elseif CalcMode == O_FIBR then
        P[period] = (H[period] + L[period]) / 2;
    elseif CalcMode == O_WOOD then
        P[period] = (H[period] + L[period] + C[period] * 2 ) / 4;
    end


    CalculateLevels(period);
    if ShowMode == O_HIST then
        local nb;
        nb = false;
        if P:hasData(period - 1) and math.abs(P[period - 1] - P[period]) > eps then
            nb = true;
        end

        for i = S1, R4, 1 do
            if show[i] and d[i] ~= 0 then
                stream[i][period] = d[i];
                if nb then
                    stream[i]:setBreak(period, true);
                end
            end
        end
    end
    if (period == source:size() - 1) then
        ShowLevels(d, period);
    end
end

-- the function is called when the async operation is finished
function AsyncOperationFinished(cookie)
    local period;

    pday = nil;
    period = P:getBookmark(1);

    if (period < 0) then
        period = 0;
    end
    loading = false;
    instance:updateFrom(period);
end

function CalculateLevels(period)
    local h, l, p, r;
    p = P[period];
    h = H[period];
    l = L[period];
    r = h - l;

    if CalcMode == O_PIVOT then
        d[R4] = p + r * 3;
        d[R3] = p + r * 2;
        d[R2] = p + r;
        d[R1] = p * 2 - l;

        d[S1] = p * 2 - h;
        d[S2] = p - r;
        d[S3] = p - r * 2;
        d[S4] = p - r * 3;
    elseif CalcMode == O_CAM then
        d[R4] = p + r * 1.1 / 2;
        d[R3] = p + r * 1.1 / 4;
        d[R2] = p + r * 1.1 / 6;
        d[R1] = p + r * 1.1 / 12;

        d[S1] = p - r * 1.1 / 12;
        d[S2] = p - r * 1.1 / 6;
        d[S3] = p - r * 1.1 / 4;
        d[S4] = p - r * 1.1 / 2;
    elseif CalcMode == O_WOOD then
        d[R4] = h + (2 * (p - l) + r);
        d[R3] = h + 2 * (p - l);
        d[R2] = p + r;
        d[R1] = p * 2 - l;

        d[S1] = p * 2 - h;
        d[S2] = p - r;
        d[S3] = l - 2 * (h - p);
        d[S4] = l - (r + 2 * (h - p));
    elseif CalcMode == O_FIB then
        d[R4] = p + 1.618 * (h - l);
        d[R3] = p + 1 * (h - l);
        d[R2] = p + 0.618 * (h - l);
        d[R1] = p + 0.382 * (h - l);

        d[S1] = p - 0.382 * (h - l);
        d[S2] = p - 0.618 * (h - l);
        d[S3] = p - 1 * (h - l);
        d[S4] = p - 1.618 * (h - l);
    elseif CalcMode == O_FLOOR then
        d[R4] = 0;
        d[R3] = h + (p - l) * 2;
        d[R2] = p + r;
        d[R1] = p * 2 - l;

        d[S1] = p * 2 - h;
        d[S2] = p - r;
        d[S3] = l - (h - p) * 2;
        d[S4] = 0;
    elseif CalcMode == O_FIBR then
        d[R4] = l + (h - l) * fibr[R4];
        d[R3] = l + (h - l) * fibr[R3];
        d[R2] = l + (h - l) * fibr[R2];
        d[R1] = l + (h - l) * fibr[R1];

        d[S1] = l + (h - l) * fibr[S1];
        d[S2] = l + (h - l) * fibr[S2];
        d[S3] = l + (h - l) * fibr[S3];
        d[S4] = l + (h - l) * fibr[S4];
    end

    return ;
end

function round(num, idp)
  local mult = 10^(idp or 0)
  return math.floor(num * mult + 0.5) / mult
end

function ShowLevels(data, period)
    local i, d1, d2;

    --d1 = source:date(0);
    d2 = source:date(period);
    d1, d2 = getCandle(d2);

    host:execute("drawLine", PID, d1, P[period], d2, P[period], clrP, styleP, widthP, "P(" .. round(P[period], source:getPrecision()) .. ")");
    if LabelLoc == O_END or LabelLoc == O_BOTH then
        host:execute("drawLabel", PID, d2, P[period], name[RP]);
    end
    if LabelLoc == O_BEG or LabelLoc == O_BOTH then
        host:execute("drawLabel", PID + 100, d1, P[period], name[RP]);
    end

    for i = S1, R4, 1 do
        if show[i] and data[i] ~= 0 then
            host:execute("drawLine", i, d1, data[i], d2, data[i], clr[i], style[i], width[i], name[i] .. "(" .. round(data[i], source:getPrecision()) .. ")");
            if LabelLoc == O_END or LabelLoc == O_BOTH then
                host:execute("drawLabel", i, d2, data[i], name[i]);
            end
            if LabelLoc == O_BEG or LabelLoc == O_BOTH then
                host:execute("drawLabel", i + 100, d1, data[i], name[i]);
            end
        else
            host:execute("removeLine", i);
            host:execute("removeLabel", i);
            host:execute("removeLabel", i + 100);
        end
    end

    if ShowMP then
        ShowMPP(0, d1, d2, d[S2], d[S3], "M0");
        ShowMPP(1, d1, d2, d[S1], d[S2], "M1");
        ShowMPP(2, d1, d2, P[period], d[S1], "M2");
        ShowMPP(3, d1, d2, P[period], d[R1], "M3");
        ShowMPP(4, d1, d2, d[R1], d[R2], "M4");
        ShowMPP(5, d1, d2, d[R2], d[R3], "M5");
    end
end

function ShowMPP(i, d1, d2, p1, p2, l)
    if p1 ~= 0 and p2 ~= 0 then
        local p = (p1 + p2) / 2;
        host:execute("drawLine", PID + 10 + i, d1, p, d2, p, clrMP, styleMP, widthMP, l .. "(" .. round(p, source:getPrecision()) .. ")");
        if LabelLoc == O_END or LabelLoc == O_BOTH then
            host:execute("drawLabel", PID + 10 + i, d2, p, l);
        end
        if LabelLoc == O_BEG or LabelLoc == O_BOTH then
            host:execute("drawLabel", PID + 110 + i, d1, p, l);
        end
    end
end


function getCandle(date)
    local cs, ce;
    cs, ce = core.getcandle("D1", date, firstHour, 0);
    local nontrading, nontradingstart, nontradingend, nontradinghours;
    nontrading, nontradingstart = core.isnontrading(cs, offset);
    if nontrading then
        -- begin of the candle belongs to nontrading
        nontradinghours = (nontradingstart + 2) - cs;
        cs = nontradingstart - nontradinghours;
    end
    nontrading, nontradingstart = core.isnontrading(ce, offset);
    if nontrading then
        -- begin of the candle belongs to nontrading
        nontradinghours = ce - nontradingstart;
        ce = nontradingstart + 2 + nontradinghours;
    end
    return cs, ce;
end

function getPrevCandle(date)
    local cs, ce;
    cs, ce = core.getcandle("D1", date, firstHour, 0);
    cs = cs - 1;
    ce = ce - 1;
    local nontrading, nontradingstart, nontradingend, nontradinghours;
    nontrading, nontradingstart = core.isnontrading(cs, offset);
    if nontrading then
        -- begin of the candle belongs to nontrading
        nontradinghours = (nontradingstart + 2) - cs;
        cs = nontradingstart - nontradinghours;
    end
    nontrading, nontradingstart = core.isnontrading(ce, offset);
    if nontrading then
        -- begin of the candle belongs to nontrading
        nontradinghours = ce - nontradingstart;
        ce = nontradingstart + 2 + nontradinghours;
    end
    return cs, ce;
end

